/* wptClipEncryptDlg.cpp - Clipboard encrypt dialog
 *	Copyright (C) 2000-2005 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptCommonCtl.h"
#include "wptGPG.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptRegistry.h"
#include "wptDlgs.h"


/* Encrypt the contents of the clipboard with the keys in @rset.
   If @always_trust is set, GPG is forced to trust any recipients.
   The context is returned in @r_ctx.
   Return value: 0 on success. */
gpgme_error_t
gpg_clip_encrypt (gpgme_key_t rset[], int always_trust, gpgme_ctx_t *r_ctx)
{
    gpgme_error_t err;
    gpgme_ctx_t ctx = NULL;
    gpgme_data_t plain = NULL;
    gpgme_data_t ciph = NULL;
    
    err = gpgme_new (&ctx);
    if (err)
        return err;
    
    gpgme_set_armor (ctx, 1);
    
    err = gpg_data_new_from_clipboard (&plain, 0);
    if (err)
	goto leave;
    err = gpgme_data_new (&ciph);
    if (err)
	goto leave;
    err = gpgme_op_encrypt (ctx, rset,
			    always_trust?GPGME_ENCRYPT_ALWAYS_TRUST : (gpgme_encrypt_flags_t)0, 
			    plain, ciph);
    *r_ctx = ctx;
    if (err)
	goto leave;
    
    gpg_data_release_and_set_clipboard (ciph, 1);
    ciph = NULL;

leave:
    if (ciph)
	gpgme_data_release (ciph);
    gpgme_data_release (plain);
    return err;
}


/* Show all invalid recipients if possible. @dlg is the 
   handle to the calling dialog and @ctx is the context
   used for encryption.
   Return value: 0 if invalid recipients exist -1 otherwise. */
static int
show_invalid_recipients (HWND dlg, gpgme_ctx_t ctx)
{
    gpgme_encrypt_result_t res;
    gpgme_invalid_key_t k;
    gpgme_key_t key;
    size_t len=0;
    char *p;

    if (!ctx)
	return -1;
    res = gpgme_op_encrypt_result (ctx);
    if (!res || !res->invalid_recipients)
	return -1;

    for (k=res->invalid_recipients; k; k = k->next) {
	get_pubkey (k->fpr, &key);
	len += (32 + 16 + strlen (key->uids->name) + 2) + 4;
    }

    p = (char *)calloc (1, len+64);
    if (!p)
	BUG (NULL);
    strcpy (p, _("Recipients unsuable for encryption:\n"));
    for (k = res->invalid_recipients; k; k = k->next) {
	get_pubkey (k->fpr, &key);
	strcat (p, key->subkeys->keyid+8);
	strcat (p, " : ");
	strcat (p, key->uids->name);
	strcat (p, "\n");
    }
    msg_box (dlg, p, _("Encryption"), MB_ERR);
    free (p);
    return 0;
}


/* Dialog procedure for the clipboard encryption. */
BOOL CALLBACK 
clip_encrypt_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    gpg_keycache_t kc;
    gpgme_key_t *rset;
    gpgme_ctx_t ctx=NULL;
    gpgme_error_t err;
    int force_trust = 0;
    int kmode = reg_prefs.keylist_mode? KEYLIST_ENCRYPT_MIN : KEYLIST_ENCRYPT;
    int n;
    
    switch( msg ) {
    case WM_INITDIALOG:
        SetWindowText (dlg, _("Encryption"));
	SetDlgItemText (dlg, IDC_ENCRYPT_FNDCMD, _("&Find"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
        kc = keycache_get_ctx( KEYCACHE_PUB );
        if( !kc )
            BUG( NULL );
        lv = keylist_load (GetDlgItem( dlg, IDC_ENCRYPT_KEYLIST ), kc, NULL,
			   kmode, KEY_SORT_USERID);
	center_window( dlg, NULL );
        SetForegroundWindow( dlg );
        return TRUE;
        
    case WM_DESTROY:
        if( lv ) {
            keylist_delete( lv );
            lv = NULL;
        }
        return FALSE;
        
    case WM_NOTIFY:
        NMHDR *notify;
        
        notify = (NMHDR *)lparam;
        if (notify && notify->code == NM_DBLCLK && 
	    notify->idFrom == IDC_ENCRYPT_KEYLIST)
            PostMessage( dlg, WM_COMMAND, MAKEWPARAM( IDOK, 0 ), 0 );
	if (notify && notify->code == LVN_COLUMNCLICK && 
	    notify->idFrom == IDC_ENCRYPT_KEYLIST ) {
	    NMLISTVIEW *p = (LPNMLISTVIEW) lparam;
	    int sortby = 0;

	    switch( p->iSubItem ) {
	    case  0: sortby = KEY_SORT_USERID; break;
	    case  1: sortby = KEY_SORT_KEYID; break;
	    case  2: sortby = KEY_SORT_LEN; break;
	    case  4: sortby = KEY_SORT_VALIDITY; break;
	    default: sortby = KEY_SORT_USERID; break;
	    }
	    keylist_sort( lv, sortby );
	}
        return TRUE;
        
    case WM_SYSCOMMAND:
        if( LOWORD( wparam ) == SC_CLOSE )
            EndDialog( dlg, TRUE );
        return FALSE;
        
    case WM_COMMAND:
        switch( LOWORD( wparam ) ) {
        case IDOK:
            rset = keylist_get_recipients (lv, &force_trust, &n);
	    if (!n) {
                msg_box (dlg, _("You must select at least one key."), _("Encryption"), MB_ERR);
                free (rset);
                return FALSE;
            }
	    err = gpg_clip_encrypt (rset, force_trust, &ctx);
            if (err) {
		if (show_invalid_recipients (dlg, ctx))
		    msg_box (dlg, gpgme_strerror (err), _("Encryption"), MB_ERR);
		if (ctx)
		    gpgme_release (ctx);
		free (rset);
                return FALSE;
            }
            else
		show_msg( dlg, 1500, _("GnuPG Status: Finished") );
            free (rset);
	    gpgme_release (ctx);
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
            EndDialog( dlg, FALSE );
            return FALSE;

	case IDC_ENCRYPT_FNDCMD:
	{
	    char tmpbuf[64];

	    n = GetDlgItemText (dlg, IDC_ENCRYPT_FIND, tmpbuf, DIM (tmpbuf)-1);
	    if (!n)
		break;
	    n = listview_find (lv, tmpbuf);
	    if (n != -1) {		
		int oldpos = listview_get_curr_pos (lv);
		listview_select_one (lv, n);
		listview_scroll (lv, oldpos, n);
	    }
	    else
		log_box (_("Encryption"), MB_ERR, _("No recipient found with '%s'"), tmpbuf);
	    break;
	}
        }
        break;
    }
    
    return FALSE;
}

